/*******************************************************************************
 * $Id: AbstractMatch.h,v 1.8 2004/02/27 23:08:55 darling Exp $
 * This file is copyright 2002-2004 Aaron Darling.  All rights reserved.
 * Please see the file called COPYING for licensing, copying, and modification
 * rights.  Redistribution of this file, in whole or in part is prohibited
 * without express permission.
 ******************************************************************************/

#ifndef _AbstractMatch_h_
#define _AbstractMatch_h_

#include "gnClone.h"
#include <iostream>
#include <vector>
#include <set>
#include <valarray>

static const gnSeqI MEM_NO_MATCH = 0;
#define MAX_SEQ_COUNT 50

/**
 * The AbstractMatch class stores the location of a matching region
 * between several sequences.
 */
class AbstractMatch : public gnClone {
public:
	AbstractMatch();
	/**
	 * Creates a new AbstractMatch.
	 * @param seq_count The total number of sequences in the alignment
	 * @param mersize The size of the mers used in the sorted mer lists.
	 * @param m_type The type of mem to create, can either be a seed or already extended.
	 * @see MatchType
	 */
	AbstractMatch(const uint32 seq_count );
	AbstractMatch(const AbstractMatch& mhe);
	AbstractMatch& operator=(const AbstractMatch& mhe);
	~AbstractMatch();
	
	virtual AbstractMatch* Clone() const = 0;
	
	/** Returns the generalized offset of this match */
	int64 Offset() const{return m_offset;}

	/** Sets the generalized offset of this match to "offset" */
	void SetOffset(int64 offset){m_offset = offset;}

	/** Returns the specified sequence's offset in this match */
	int64 Offset( uint32 seqI ) const{ return m_offsets[ seqI ]; }

	/** Returns the length of this match */
	virtual gnSeqI Length( uint seqI ) const = 0;

	/** Sets the length of this match to @param len */
	virtual void SetLength( gnSeqI len, uint seqI ) = 0;

	/** Returns the start coordinate of this match in sequence @param startI */
	int64 Start(uint32 startI) const{return m_start[startI];}

	/** Sets the start in sequence @param seqI of this match to @param startI */
	void SetStart(uint32 seqI, int64 startI);

	/** Returns the start coordinate of this match in sequence @param startI */
	int64 operator[](uint32 startI) const{return m_start[startI];}

	/** Returns the start coordinate of this match in sequence @param startI */
	int64& operator[](uint32 startI) {return m_start[startI];}

	/** Returns the last coordinate of this match in sequence @param endI */
	virtual int64 End(uint32 endI) const = 0;

	/** Returns the multiplicity of the match.  e.g. the number of sequences this match occurs in */
	uint32 Multiplicity() const{return m_multiplicity;}

	/** Returns the number of sequences in the alignment which contains this match */
	uint32 SeqCount() const{return m_seq_count;}

	/** Returns the index of the first sequence this match occurs in */
	virtual uint32 FirstStart() const;

	/** Returns a numerical representation of which sequences are matched */
	uint64 MatchNumber() const {return m_matchnumber;}

//	/** Returns a numerical representation of which sequences are matched */
//	virtual valarray<bool> MatchNumber() const {return m_matchnumber;}
	
	/**
	 * Inverts the coordinates of this mem.
	 */
	virtual void Invert() = 0;
	
	/**
	 * Calculates the generalized offset and other bookkeeping information
	 * for this mem.  This should <b>always</b> be called after changing the start
	 * positions of the mem.
	 */
	virtual void CalculateOffset();
	
	/**
	 * Return the difference between the end of this mem and the start
	 * of mhe in the sequence specified by seqI
	 * @param mhe The mem to compare to.
	 * @param seqI The index of the sequence to compare
	 * @return The gap size.
	 */
	virtual int64 GapSize(const AbstractMatch& mhe, uint32 seqI) = 0;
	
	//warning:  none of the following do bounds checking.
	/** 
	 * Removes the first <code>crop_amount</code> base pairs from the beginning of the match.
	 */
	virtual void CropStart(gnSeqI crop_amount) = 0;
	/** 
	 * Removes the last <code>crop_amount</code> base pairs from the end of the match.
	 */
	virtual void CropEnd(gnSeqI crop_amount) = 0;

	
	/**
	 * Writes the location of this match to the specified output stream (e.g. cout).
	 */
//	friend std::ostream& operator<<(std::ostream& os, const AbstractMatch& mhe) = 0; //write to source.

protected:

	int64 m_offset;
	uint m_seq_count;
	int64 m_start[ MAX_SEQ_COUNT ];
	uint32 m_multiplicity;
	uint32 m_firstStart;
//	valarray<bool> m_matchnumber;
	uint64 m_matchnumber;
	int64 m_offsets[ MAX_SEQ_COUNT ];
};


inline
void AbstractMatch::SetStart(uint32 seqI, int64 startI){
	m_start[seqI] = startI;
}

class MatchStartComparator {
public:
	MatchStartComparator( unsigned seq = 0 ){
		m_seq = seq;
	}
	MatchStartComparator( MatchStartComparator& msc ){
		m_seq = msc.m_seq;
	}
	// TODO??  make this do a wraparound comparison if all is equal?
	boolean operator()(const AbstractMatch* a, const AbstractMatch* b) const{
		int32 start_diff = max( a->FirstStart(), m_seq ) - max( b->FirstStart(), m_seq );
		if(start_diff == 0){
			uint32 m_count = a->SeqCount();
			m_count = m_count <= b->SeqCount() ? m_count : b->SeqCount();
			for(uint32 seqI = m_seq; seqI < m_count; seqI++){
				int64 a_start = a->Start( seqI ), b_start = b->Start( seqI );
				if(a_start < 0)
					a_start = -a_start + a->Length( seqI );
				if(b_start < 0)
					b_start = -b_start + b->Length( seqI );
				int64 diff = a_start - b_start;
				if(a_start == (int64)MEM_NO_MATCH || b_start == (int64)MEM_NO_MATCH)
					continue;
				else if(diff == 0)
					continue;
				else
					return diff < 0;
			}
		}
		return start_diff < 0;
	}
private:
	unsigned m_seq;
};

class SingleStartComparator {
public:
	SingleStartComparator( unsigned seq = 0 ){
		m_seq = seq;
	}
	SingleStartComparator( SingleStartComparator& msc ){
		m_seq = msc.m_seq;
	}
	/**
	 * Compare on only one sequence.  Undefined matches are less than defined matches
	 */
	boolean operator()(const AbstractMatch* a, const AbstractMatch* b) const{
		int64 a_start = a->Start( m_seq ), b_start = b->Start( m_seq );
		if( a_start == (int64)MEM_NO_MATCH || b_start == (int64)MEM_NO_MATCH ){
			if( b_start != (int64)MEM_NO_MATCH )
				return true;
			return false;
		}
		if(a_start < 0)
			a_start = -a_start;
		if(b_start < 0)
			b_start = -b_start;

		int64 diff = a_start - b_start;
		return diff < 0;
	}
private:
	unsigned m_seq;
};


#endif // _AbstractMatch_h_
