/*******************************************************************************
 * $Id: ClustalInterface.h,v 1.12 2004/04/19 23:10:50 darling Exp $
 * This file is copyright 2002-2004 Aaron Darling.  All rights reserved.
 * Please see the file called COPYING for licensing, copying, and modification
 * rights.  Redistribution of this file, in whole or in part is prohibited
 * without express permission.
 ******************************************************************************/

#ifndef _GappedAlignment_h_
#define _GappedAlignment_h_

#include "libMems/NumericMatrix.h"
#include "gn/gnFilter.h"
#include "libMems/Match.h"
#include "gn/gnSequence.h"

class GappedAlignment : public AbstractMatch {
public:
	GappedAlignment();
	GappedAlignment( uint seq_count, gnSeqI align_length );
	GappedAlignment( const GappedAlignment& cr );
	~GappedAlignment();
	GappedAlignment& operator=( const GappedAlignment& cr );
	
	GappedAlignment* Clone() const { return new GappedAlignment( *this ); }
	
	void setAlignment( vector< string >& seq_align, double score );

	double getScore() const;
	string getColumn( gnSeqI colI ) const;
	double getIdentity( uint seqI, uint seqJ ) const;

	vector< string > align_matrix;

	/**
	 * Writes this GappedAlignment to the specified output stream (e.g. cout).
	 */
	friend std::ostream& operator<<(std::ostream& os, const GappedAlignment& cr); //write to source.

	/**
	 * Reads a GappedAlignment from the specified input stream (e.g. cin).
	 */
	friend std::istream& operator>>(std::istream& is, const GappedAlignment& cr); //read from source

	gnSeqI align_length;
	double score;

	// Inherited methods from AbstractMatch:
	virtual gnSeqI Length( uint seqI = UINT_MAX ) const; 
	virtual void SetLength( gnSeqI len, uint seqI ) { length[ seqI ] = len; }
	virtual int64 End(uint32 endI) const;
	virtual void Invert();
	virtual int64 GapSize(const AbstractMatch& mhe, uint32 seqI);
	virtual void CropStart(gnSeqI crop_amount);
	virtual void CropEnd(gnSeqI crop_amount);

protected:
	vector< int64 > length;
	NumericMatrix< double > identity;
};

inline
gnSeqI GappedAlignment::Length( uint seqI ) const { 
	if( seqI == UINT_MAX )
		return align_length;
	return length[ seqI ]; 
}

inline
int64 GappedAlignment::End(uint32 endI) const {
	if( m_start[ endI ] > 0 )
		return m_start[endI] + align_length - 1;
	return m_start[ endI ];
}

inline
void GappedAlignment::Invert(){
	const gnFilter* rc_filter = gnFilter::DNAComplementFilter();
	for(uint32 startI = 0; startI < SeqCount(); startI++){
		m_start[startI] = -m_start[startI];
		rc_filter->ReverseFilter( align_matrix[ startI ] );
		
	}
}

inline
void GappedAlignment::CropStart(gnSeqI crop_amount){
	if( crop_amount > align_length )
		Throw_gnEx( SeqIndexOutOfBounds() );
	for( uint32 i=0; i < SeqCount(); i++ ){
		align_matrix[ i ] = align_matrix[ i ].substr( crop_amount );
		if( m_start[i] > 0 )
			m_start[i] += crop_amount;
	}
	align_length -= crop_amount;
}

inline
void GappedAlignment::CropEnd(gnSeqI crop_amount){
	if( crop_amount > align_length )
		Throw_gnEx( SeqIndexOutOfBounds() );
	align_length -= crop_amount;
	for( uint32 i=0; i < SeqCount(); i++ ){
		align_matrix[ i ] = align_matrix[ i ].substr( 0, align_length );
		if( m_start[i]  < 0 )
			m_start[i] -= crop_amount;
	}
}

inline
int64 GappedAlignment::GapSize(const AbstractMatch& mhe, uint32 seqI){
	cerr << "GappedAlignment::GapSize not fully implemented\n";
	int64 start_1 = m_start[seqI];
	int64 start_2 = mhe.Start( seqI );

	//check for match existence
	//and for similar match direction
	if((start_2 < 0 && start_1 > 0) ||
		(start_2 > 0 && start_1 < 0) ||
		(start_2 == 0 && start_1 != 0) ||
		(start_2 != 0 && start_1 == 0))
		Throw_gnEx(SeqIndexOutOfBounds());
	
	// this code can probably be simplified
	if(start_2 > 0)
		return start_2 - (start_1 + (int64)align_length);
	if(start_2 < 0)
		return -start_1 - (-start_2 + mhe.Length(seqI));
	return 0;
}

#endif // _GappedAlignment_h_

