/*******************************************************************************
 * $Id: IntervalList.cpp,v 1.12 2004/03/01 02:40:08 darling Exp $
 * This file is copyright 2002-2004 Aaron Darling.  All rights reserved.
 * Please see the file called COPYING for licensing, copying, and modification
 * rights.  Redistribution of this file, in whole or in part is prohibited
 * without express permission.
 ******************************************************************************/

#include "libMems/IntervalList.h"
#include "libMems/DNAFileSML.h"
#include "libMems/MemHash.h"
#include <map>
#include <sstream>


IntervalList::IntervalList( const IntervalList& ml ){
	*this = ml;
}

IntervalList& IntervalList::operator=( const IntervalList& ml ){
	vector< Interval >::operator=( ml );
	seq_filename = ml.seq_filename;
	seq_table = ml.seq_table;
	return *this;
}

void IntervalList::Clear() {
	for( uint seqI = 0; seqI < seq_table.size(); seqI++ ){
		if( seq_table[ seqI ] != NULL )
			delete seq_table[ seqI ];
	}
	seq_filename.clear();
	clear();
}

void IntervalList::ReadList(istream& match_file){
	string tag;
	gnSeqI len;
	int64 start;
	unsigned int seq_count;
	uint seqI;
	
	match_file >> tag;	//format version tag
	if( tag != "FormatVersion" ){
		Throw_gnEx(InvalidFileFormat());
	}
	match_file >> tag;	//format version
	if( tag != "4" ){
		Throw_gnEx(InvalidFileFormat());
	}
	match_file >> tag;	//sequence count tag
	if( tag != "SequenceCount" ){
		Throw_gnEx(InvalidFileFormat());
	}
	match_file >> seq_count;	//sequence count
	if(seq_count < 2){
		Throw_gnEx(InvalidFileFormat());
	}
	
	vector< string > alignment;
	// read the sequence file names and lengths
	for( seqI = 0; seqI < seq_count; seqI++ ){
		match_file >> tag;	// name tag
		match_file >> tag;	// name
		seq_filename.push_back(tag);
//		try{
//			gnSequence *new_seq = new gnSequence();
//			new_seq->LoadSource(tag);
//			seq_table.push_back( new_seq );
//		}catch( gnException& gne );
		match_file >> tag;	// length tag
		match_file >> tag;	// length

		alignment.push_back( "" );	// initialize alignment vector
	}
	uint interval_count;
	match_file >> tag;	// interval count tag
	match_file >> interval_count;	// interval count
	
	
	// read the matches
	string cur_line;
	Interval* cur_iv = NULL;
	boolean clustal_match;

	// use a match allocator
	SlotAllocator< Match >& match_allocator = SlotAllocator< Match >::GetSlotAllocator();
	
	while( getline( match_file, cur_line ) ){
		if( cur_line.find( "Interval" ) != string::npos ){
			// end the old interval
			if( cur_iv != NULL )
				cur_iv->CalculateOffset();
			// start a new interval
			Interval iv( seq_count );
			push_back( iv );
			cur_iv = &(*this)[ size() - 1 ];
			continue;
		}
		if( cur_iv == NULL )
			continue;
		if( cur_line.length() == 0 )
			continue;
		
		clustal_match = false;
		if( cur_line == "GappedAlignment" ){
			clustal_match = true;
			getline( match_file, cur_line );

			stringstream line_stream( cur_line );
			line_stream >> len;
			GappedAlignment* cr = new GappedAlignment( seq_count, len );
			
			for( seqI = 0; seqI < seq_count; seqI++ ){
				line_stream >> start;
				cr->SetStart( seqI, start );
				getline( match_file, alignment[ seqI ] );
				int64 seq_len = 0;
				for( uint charI = 0; charI < alignment[ seqI ].length(); charI++ )
					if( alignment[ seqI ][ charI ] != '-' )
						seq_len++;
				cr->SetLength( seq_len, seqI );
			}
			cr->setAlignment( alignment, 0 );
			cur_iv->add( cr );
		}
		else{

			Match* mhe = match_allocator.Allocate();
			*mhe = Match( seq_count, 0 );
			stringstream line_stream( cur_line );
			
			line_stream >> len;
			mhe->SetLength(len);

			for( seqI = 0; seqI < seq_count; seqI++){
				line_stream >> start;
				mhe->SetStart(seqI, start);
			}
		
			mhe->CalculateOffset();
			cur_iv->add( mhe );
		}
	}
	if( cur_iv != NULL )
		cur_iv->CalculateOffset();
	if( interval_count != size() ){
		Throw_gnEx(InvalidFileFormat());
	}
	
}

void IntervalList::WriteList(ostream& match_file) const{

	unsigned int seq_count = seq_table.size();
	uint seqI;
	
	match_file << "FormatVersion" << '\t' << 4 << "\n";
	match_file << "SequenceCount" << '\t' << seq_count << "\n";
	for(seqI = 0; seqI < seq_count; seqI++){
		match_file << "Sequence" << seqI << "File" << '\t';
		if( seq_filename.size() > seqI )
			match_file << seq_filename[seqI];
		else
			match_file << "null";
		match_file << "\n";
		match_file << "Sequence" << seqI << "Length" << '\t';
		if( seq_table.size() > seqI )
			match_file << seq_table[seqI]->length();
		else
			match_file << "0";
		match_file << "\n";
	}

	match_file << "IntervalCount" << '\t' << size() << endl;
	
	for( uint ivI = 0; ivI < size(); ivI++ ){
		match_file << "Interval " << ivI << endl;
		for( uint matchI = 0; matchI < (*this)[ ivI ].matches.size(); matchI++ ){
			const AbstractMatch* m = (*this)[ ivI ].matches[ matchI ];
			const GappedAlignment* cr = dynamic_cast< const GappedAlignment* >( m );
			const Match* match = dynamic_cast< const Match* >( m );
			if( match != NULL ){
				match_file << *match << endl;
			}
			else if( cr != NULL ){
				match_file << "GappedAlignment\n";
				match_file << cr->Length();
				for( seqI = 0; seqI < seq_count; seqI++ )
					match_file << '\t' << cr->Start( seqI );
				match_file << endl;
				for( seqI = 0; seqI < seq_count; seqI++ )
					match_file << cr->align_matrix[ seqI ] << endl;
			}
		}
		match_file << endl;
	}
}

void IntervalList::WriteStandardAlignment( ostream& out_file ) const {
	if( size() == 0 )
		return;

	unsigned int seq_count = seq_table.size();

	for( uint ivI = 0; ivI < size(); ivI++ ){
		if( (*this)[ ivI ].matches.size() == 0 ){
			continue;
		}
		
		gnAlignedSequences gnas;
		(*this)[ ivI ].GetAlignedSequences( gnas, seq_table );
		for( uint seqI = 0; seqI < seq_count; seqI++ ){
			out_file << "> " << seqI + 1 << ":";
			int64 startI = (*this)[ ivI ].Start( seqI );
			gnSeqI length = (*this)[ ivI ].Length( seqI );
			if( startI > 0 ){
				out_file << abs( startI ) << "-" << abs( startI ) + length - 1 << " + ";
			}else if( startI == 0 ){
				out_file  << "0-0 + ";
			}else{
				out_file << abs( startI ) << "-" << abs( startI ) + length - 1 << " - ";
			}
			out_file << endl;
			gnSeqI cur_pos = 0;
			for( ; cur_pos < gnas.sequences[ seqI ].length(); cur_pos += 80 ){
				gnSeqI cur_len = cur_pos + 80 < gnas.sequences[ seqI ].length() ? 80 : gnas.sequences[ seqI ].length() - cur_pos;
				out_file.write( gnas.sequences[ seqI ].data() + cur_pos, cur_len );
				out_file << endl;
			}
		}
		out_file << "=" << endl;
	}
	
}

void IntervalList::ReadStandardAlignment( istream& in_stream ) {
	uint seq_count = 0;
	string cur_line;
	getline( in_stream, cur_line );
	uint seqI = 0;
	vector< gnSeqI > lengths;
	GappedAlignment cr;
	while( true ){
		
		// read and parse the def. line
		stringstream line_str( cur_line );
		getline( line_str, cur_line, '>' );
		getline( line_str, cur_line, ':' );
		// take off leading whitespace
		stringstream parse_str( cur_line );
		string name;
		parse_str >> name;
		gnSeqI start, stop;
		getline( line_str, cur_line, '-' );
		parse_str.clear();
		parse_str.str( cur_line );
		parse_str >> start;
		line_str >> stop;
		string strand;
		line_str >> strand;

		// read and parse the sequence
		cr.align_matrix.push_back( "" );
		gnSeqI chars = 0;
		while( getline( in_stream, cur_line ) ){
			if( (cur_line[ 0 ] == '>' ) || (cur_line[ 0 ] == '=' ))
				break;
			for( uint charI = 0; charI < cur_line.length(); charI++ )
				if( cur_line[ charI ] != '-' )
					chars++;
			cr.align_matrix[ seqI ] += cur_line;
		}
		lengths.push_back( chars );
// temporary workaround for file format inconsistency
		if( strand == "+" )
			cr.SetStart( seqI, start );
		else if( start < stop )
			cr.SetStart( seqI, -start );
		else
			cr.SetStart( seqI, -stop );
			
		seqI++;
		if( cur_line.size() == 0 )
			break;
		// did we finish an aligned region?
		if( cur_line[ 0 ] != '>' ){
			GappedAlignment *new_cr = new GappedAlignment( cr.align_matrix.size(), cr.align_matrix[ 0 ].length() );
			for( uint seqJ = 0; seqJ < seqI; seqJ++ ){
				new_cr->SetStart( seqJ, cr.Start( seqJ ) );
				new_cr->SetLength( lengths[ seqJ ], seqJ );
			}
			new_cr->align_matrix = cr.align_matrix;
			lengths.clear();
			new_cr->CalculateOffset();
			if( seq_count == 0 )
				seq_count = seqI;
			else if( seq_count != seqI )
				ErrorMsg( "Error parsing XMFA alignment file\n" );
			seqI = 0;
			Interval iv( seq_count, cr.align_matrix[ 0 ].length() );
			for( uint seqJ = 0; seqJ < seq_count; seqJ++ ){
				iv.SetStart( seqJ, cr.Start( seqJ ) );
			}
			push_back( iv );
			(*this)[ size() - 1 ].add( new_cr );
			(*this)[ size() - 1 ].CalculateOffset();
			cr.align_matrix.clear();	// reset cr for next interval

			// bail out on EOF or corruption
			if( cur_line[ 0 ] != '=' )
				break;
			// otherwise read up to the next def. line
			while( getline( in_stream, cur_line ) ){
				if( cur_line[ 0 ] == '>' )
					break;
			}
			if( cur_line[ 0 ] != '>' )
				break;
		}
	}
}

void IntervalList::WriteAlignedSequences(ostream& match_file) const{

	unsigned int seq_count = seq_table.size();
	uint seqI;
	
	match_file << "mauveAligner data\n";
	match_file << "FormatVersion" << '\t' << 5 << "\n";
	match_file << "SequenceCount" << '\t' << seq_count << "\n";
	for(seqI = 0; seqI < seq_count; seqI++){
		match_file << "Sequence" << seqI << "File" << '\t';
		if( seq_filename.size() > seqI )
			match_file << seq_filename[seqI];
		else
			match_file << "null";
		match_file << "\n";
		match_file << "Sequence" << seqI << "Length" << '\t';
		if( seq_table.size() > seqI )
			match_file << seq_table[seqI]->length();
		else
			match_file << "0";
		match_file << "\n";
	}

	match_file << "AlignmentCount" << '\t' << size() << endl;

	if( size() == 0 )
		return;
	
	for( uint ivI = 0; ivI < size(); ivI++ ){
		gnAlignedSequences gnas;

		if( (*this)[ ivI ].matches.size() == 0 ){
			match_file << "0";
			for( seqI = 0; seqI < seq_count; seqI++ )
				match_file << "\t0";
			match_file << endl;
			for( seqI = 0; seqI < seq_count; seqI++ )
				match_file << endl;
			match_file << endl;
			continue;
		}
		
		match_file << (*this)[ ivI ].matches[ 0 ]->Length(0);
		for( seqI = 0; seqI < seq_count; seqI++ )
			match_file << '\t' << (*this)[ ivI ].matches[ 0 ]->Start( seqI );
		match_file << endl;
		(*this)[ ivI ].GetAlignedSequences( gnas, seq_table );
		for( seqI = 0; seqI < seq_count; seqI++ )
			match_file << gnas.sequences[ seqI ] << endl;
		match_file << endl;
	}
	
}
